function [l, L, q, u, costs, data] = RSOC(task, model, RSOCSettings)
%function [l, L, q, u, costs] = RSOC(task, model) 
% **********************************************************
% *
% * Robust Policy Updates for Stochastic Optimal Control
% *   E. Rueckert, M. Mindt, J. Peters and G. Neumann (HUMANOIDS 2014)
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert@ias.tu-darmstadt.de
% *
% **********************************************************
% Inputs are task and robot model desctriptions (as structures with
% function pointers, see template files), optional are algorithm
% specific parameters
% Outputs: 
%   a) time varying feedback controller: u(:,t) = L(:,:,t)*q(:,t) + l(:,t)
%   b) q ... joint angles and control torques u
%   c) costs ... vector of the costs of the internal optimization process,
%   where the structure data contains the internal generated trajectories

if nargin < 2
    error('Call this function with at least two arguments!');
end
if nargin < 3
    RSOCSettings = getDefaultRSOCSettings();
end

fnDyn = model.fnDyn;
fnCost = task.fnCost;

[maxIts, dt, n, szQ, szU, flgPrintCost, q0, Q, q, u, theta, maxTimeSliceIts, flgEarlyStop] = getConstants(model, RSOCSettings);
[a, A, Qd, s, S, v, V, r, R, b, B, Bm, H, h, qBelief, l, L] = initVariables(szQ, szU, n, q0);

qOld = q;


its = 0;
ct = 0;
ctReal = 0;
for k = 2:n 
    ct = ct + fnCost(q(:,k), u(:,k), k, qOld(:,k));
end

if flgPrintCost
    disp(['RSOC iteration [' num2str(its) '] costs: ' num2str(ct)]);
end
oldCosts = ct; 
    


costs = nan(maxIts,1);

flgIsKinematicSystem = 0;
[f, f_q, f_u] = fnDyn(q(:,1), u(:,1));
a(:,1) = (f - f_q*q(:,1) - f_u*u(:,1))*dt;
A(:,:,1) = eye(szQ) + dt * f_q;
Bm(:,:,1) = dt * f_u;
if sum(diag(A(:,:,1))) == size(A,1)
    flgIsKinematicSystem = 1;
    for k=2:n
        a(:,k) = a(:,1);
        A(:,:,k) = A(:,:,1);
        Bm(:,:,k) = Bm(:,:,1);
    end
end

% keyboard
while its < maxIts
    
    if nargout > 5
       data{its+1}.costs = ct;
       data{its+1}.q = q;
       data{its+1}.u = u;
       data{its+1}.l = l;
       data{its+1}.L = L;
    end
   
    costs(its+1,1) = ct; 
    qOld = q;
     
    for k=1:n
        flgSliceConv = 0;
        itsTimeSlice = 0;

        while ~flgSliceConv && (itsTimeSlice < maxTimeSliceIts)
            
            %linearize model 
            if ~flgIsKinematicSystem
                [f, f_q, f_u] = fnDyn(q(:,k), u(:,k));
                a(:,k) = (f - f_q*q(:,k) - f_u*u(:,k))*dt;
                A(:,:,k) = eye(szQ) + dt * f_q;
                Bm(:,:,k) = dt * f_u;
            end

            %compute costs (conditionals)
            [ct, H(:,:,k), h(:,k), R(:,:,k), r(:,k)] = fnCost(q(:,k), u(:,k), k, qOld(:,k)); 
            Qtmp = Q + Bm(:,:,k)*H(:,:,k)^-1*Bm(:,:,k)';
            Qd(:,:,k) = inv(Qtmp); 

            %fwd message and intermediate
            if k > 1
                Sd = S(:,:,k-1) +  R(:,:,k-1); 
                sd = s(:,k-1) + r(:,k-1);   
                AInv = A(:,:,k-1)^-1;
                Ks = AInv'*Sd*(Sd + AInv'*Qd(:,:,k-1)*AInv)^-1;
                S(:,:,k) = (AInv' - Ks)*Sd*AInv;
                s(:,k) = (AInv' - Ks)*(sd + Sd*AInv*(a(:,k-1)+ Bm(:,:,k-1)*H(:,:,k-1)^-1*h(:,k-1)));  
            end
            
            %bwd message  
            if k < n
                Vd = V(:,:,k+1) +  R(:,:,k+1);
                vd = v(:,k+1) + r(:,k+1);
                Kv = A(:,:,k)'*Vd*(Vd + Qd(:,:,k))^-1;
                V(:,:,k) = (A(:,:,k)' - Kv)*Vd*A(:,:,k);
                v(:,k) = (A(:,:,k)' - Kv)*(vd - Vd*(a(:,k)+ Bm(:,:,k)*H(:,:,k)^-1*h(:,k)));
            end

            %compute belief
            b(:,k) = s(:,k) + v(:,k) + r(:,k); 
            B(:,:,k) = S(:,:,k) + V(:,:,k) + R(:,:,k);
            qBelief(:,k) = B(:,:,k)^-1*b(:,k);
            
%             qOld(:,k) = q(:,k);
            %q(:,k) = (1-alpha)*q(:,k) + alpha*qBelief(:,k);

            if sqrt((q(:,k) - qBelief(:,k))'*(q(:,k) - qBelief(:,k))) < theta
                flgSliceConv = 1;
                %keyboard
            end
            itsTimeSlice = itsTimeSlice + 1;
        end
    end    
         
    for k=n:-1:1 
        flgSliceConv = 0;
        itsTimeSlice = 0;

        while ~flgSliceConv && (itsTimeSlice < maxTimeSliceIts)
         
            %linearize model 
            [f, f_q, f_u] = fnDyn(q(:,k), u(:,k));
            a(:,k) = (f - f_q*q(:,k) - f_u*u(:,k))*dt;
            A(:,:,k) = eye(szQ) + dt * f_q;
            Bm(:,:,k) = dt * f_u;

            %compute costs (conditionals)
            [ct, H(:,:,k), h(:,k), R(:,:,k), r(:,k)] = fnCost(q(:,k), u(:,k), k, qOld(:,k)); 
            Qtmp = Q + Bm(:,:,k)*H(:,:,k)^-1*Bm(:,:,k)';
            Qd(:,:,k) = inv(Qtmp);% + 1e-8*eye(szQ)); 

            %fwd message and intermediate
            if k > 1
                Sd = S(:,:,k-1) +  R(:,:,k-1); 
                sd = s(:,k-1) + r(:,k-1);   
                AInv = A(:,:,k-1)^-1;
                Ks = AInv'*Sd*(Sd + AInv'*Qd(:,:,k-1)*AInv)^-1;
                S(:,:,k) = (AInv' - Ks)*Sd*AInv;
                s(:,k) = (AInv' - Ks)*(sd + Sd*AInv*(a(:,k-1)+ Bm(:,:,k-1)*H(:,:,k-1)^-1*h(:,k-1)));  
            end
            
            %bwd message  
            if k < n
                Vd = V(:,:,k+1) +  R(:,:,k+1);
                vd = v(:,k+1) + r(:,k+1);
                Kv = A(:,:,k)'*Vd*(Vd + Qd(:,:,k))^-1;
                V(:,:,k) = (A(:,:,k)' - Kv)*Vd*A(:,:,k);
                v(:,k) = (A(:,:,k)' - Kv)*(vd - Vd*(a(:,k)+ Bm(:,:,k)*H(:,:,k)^-1*h(:,k)));
            end

            %compute belief
            b(:,k) = s(:,k) + v(:,k) + r(:,k); 
            B(:,:,k) = S(:,:,k) + V(:,:,k) + R(:,:,k);
            qBelief(:,k) = B(:,:,k)^-1*b(:,k);
            
%             qOld(:,k) = q(:,k);
            %q(:,k) = (1-alpha)*q(:,k) + alpha*qBelief(:,k);

            if sqrt((q(:,k) - qBelief(:,k))'*(q(:,k) - qBelief(:,k))) < theta
                flgSliceConv = 1;
            end
            itsTimeSlice = itsTimeSlice + 1;
        end
    end 
    
    %compute controls
    %u = diff(q,1,2); %only possible within kinematic planning tasks
    %OR 
    q(:, 1) = q0;
    for k = 1:n-1
        Vd = V(:,:,k+1) + R(:,:,k+1);
        Vdstar = (Q + Vd^-1)^-1;
        vd = v(:,k+1) + r(:,k+1); 

        %feedback controller 
        MInv = (Bm(:,:,k)'*Vdstar*Bm(:,:,k)+H(:,:,k))^-1;
        G = MInv * Bm(:,:,k)'*Vdstar; 
        l(:,k) = (G*Vd^-1*vd + MInv*h(:,k) - G*a(:,k));
        L(:,:,k) = -G*A(:,:,k); 

        u(:,k) = l(:,k) + L(:,:,k)*q(:,k);

        q(:,k+1) = A(:,:,k)*q(:,k) + a(:,k) + Bm(:,:,k)*u(:,k);

    end
    if nargin > 4 && ~isempty(fnPlotResults) 
        fnPlotResults(q);
    end

    %Evaluate costs
    its = its + 1; 
    ct = 0;
    for k = 2:n 
        ct = ct + (fnCost(q(:,k),  u(:,k), k, qOld(:,k)));
    end
    if flgPrintCost
        disp(['RSOC iteration [' num2str(its) '] costs: ' num2str(ct) ' conv.: ' num2str(abs(oldCosts - ct))]);
    end
    
         
    %check overall convergence
    if flgEarlyStop && (abs(oldCosts - ct) < 1e-5) % && its > 3
      %costs = costs(1:its,1);  
      break; 
    end
    oldCosts = ct;
end
    
end

function [maxIts, dt, n, szQ, szU, flgPrintCost, q0, Q, q, u, theta, maxTimeSliceIts, flgEarlyStop] = getConstants(model, RSOCSettings)
    
    maxIts = RSOCSettings.maxIts;
    flgPrintCost = RSOCSettings.flgPrintCost;
    theta = RSOCSettings.theta;
    maxTimeSliceIts = RSOCSettings.maxTimeSliceIts;
    flgEarlyStop = RSOCSettings.flgEarlyStop;
    
    dt = model.dt;
    n = model.n;
    szQ = model.szQ;
    szU = model.szU;
    q0 = model.q0;
    Q = model.Q;
    q = model.qInit;
    u = model.uInit;
    
end

function [a, A, Qd, s, S, v, V, r, R, b, B, Bm, H, h, qBelief, l, L] = initVariables(szQ, szU, n, q0)
    a = zeros(szQ,n);
    A = zeros(szQ, szQ ,n);
    Qd = zeros(szQ, szQ ,n);
     
    s = zeros(szQ,n);
    S = zeros(szQ, szQ ,n);
    v = zeros(szQ,n);
    V = zeros(szQ, szQ ,n);
    r = zeros(szQ,n);
    R = zeros(szQ, szQ ,n);
    
    b = zeros(szQ,n);
    B = zeros(szQ, szQ ,n);

    S(:,:,1) = eye(szQ).*1e+10;
%     S(:,:,1) = eye(szQ).*1e+3;
    s(:,1) = S(:,:,1)*q0;
    
    Bm = zeros(szQ, szU ,n);
    H = zeros(szU, szU ,n);
    h = zeros(szU,n);
    qBelief = zeros(szQ,n);
    L = zeros(szU, szQ ,n);
    l = zeros(szU ,n);
end