function model = Pendulum(flgUseAnalyticDynamics)
%function model = Pendulum(flgUseAnalyticDynamics)
% **********************************************************
% *
% * Robust Policy Updates for Stochastic Optimal Control
% *   E. Rueckert, M. Mindt, J. Peters and G. Neumann (HUMANOIDS 2014)
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert@ias.tu-darmstadt.de
% *
% **********************************************************
% This function generates a structure of the robot model. 

model.dt = 50e-3;                           %50ms
model.n = 500;                              %25 seconds
model.szQ = 2;                              %joint angle phi and velocity
model.szU = 1;                              %acceleration
model.q0 = [0.001;0];
model.Q = 1e-12 * diag([ones(model.szQ/2,1)', 10*ones(model.szQ/2,1)']);              %model noise
model.qInit = repmat(model.q0,1,model.n);   %initial joint angles prior
model.uInit = zeros(1,model.n);             %initial torques, in fact n-1 would be correct but the n-th element is assumed to be zero such that C(q(T)) = C(q(T),u(T))
model.l = 1;                                %pendulum link length

if flgUseAnalyticDynamics
    model.fnDyn = @PendulumDynamics;
else
    model.fnDyn = @PseudoDynamics;
end

end


function [qdot, qdotq, qdotu] = PseudoDynamics(q, u)

    szQ = size(q,1);
    vIds = szQ/2+1:szQ;
    
    qdot = [q(vIds); u];
    
    if nargout > 1
        szU = size(u,1);
        qdotq = zeros(szQ, szQ);
        for i=1:szQ/2
            qdotq(i,vIds(i)) = 1;
        end
        qdotu = [zeros(szU, szU); diag(ones(1,szU))];
    end
    
end

function [qdot, qdotq, qdotu] = PendulumDynamics(q,u)

    qdot = pendulumDyn(q,u);
    
    [ mu,l,m,g ] = getConstants;
    if nargout > 1
        qdotq = [[0; -g/l*cos(q(1))], [1; -mu/(m*l*l)]];
        qdotu = [0 ; 1/(m*l*l)];
    end
    
end

function qdot = pendulumDyn(q,u)

    [ mu,l,m,g ] = getConstants;

    qdot = zeros(2,1);
    
    qdot(1) = q(2);
    qdot(2) = (-mu * q(2) - m*g*l*sin(q(1)) + u)/(m*l^2);

end

function [ mu,l,m,g ] = getConstants

    mu  = 0.2;%1;   % friction term
    l   = 1;    % length of the pendulum
    m   = 1;    % mass of the pendulum
    g   = 9.81; % gravity

end
