function task = PendulumSwingUpTask(alpha, dt, linkLength, numTimeSteps)
%function task = PendulumSwingUpTask(alpha)
% **********************************************************
% *
% * Robust Policy Updates for Stochastic Optimal Control
% *   E. Rueckert, M. Mindt, J. Peters and G. Neumann (HUMANOIDS 2014)
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert@ias.tu-darmstadt.de
% *
% **********************************************************
% This function generates a structure of the planning task. 
    

    task.fnCost = @(q_,u_,k_,qOld_)swingUpTask(q_,u_,k_,qOld_, alpha, dt, linkLength, numTimeSteps);
end


function [ct, H, h, R, r] = swingUpTask(q, u, k, qOld, alpha, dt, l, n)


    xTarget = [0; 1];           %x-,y-coordinate of the ende effector
    wuInLogSpace = -1;
    wxInLogSpace = [3; 4];
    
    %activate task space constraints only in the last 50% of the movement
    if k > 0.6*n && mod(k,10) == 0
        flgActivateTSCosts = 1;
    else
        flgActivateTSCosts = 0;
    end
        
    
    szQ = size(q,1);
    
    %Control Costs
    H = 10^wuInLogSpace;
    h = 0;
    
    %Task Space Costs
    if flgActivateTSCosts
        C = diag(10.^wxInLogSpace);
        qvel = q;
        posIds = 1:szQ/2;
        velIds = szQ/2+1:szQ;
        qvel(posIds) = qvel(posIds) + dt*q(velIds);
        [numEFF, EF, J] = cFnJacobianPendulum(q, l);
        [numEFF, EFvel, Jvel] = cFnJacobianPendulum(qvel, l);
        Jfull = [J dt*Jvel];
        RTaskSpace = Jfull'*C*Jfull;
        distanceEF = Minus_pi_to_pi_EF( xTarget - EF );   
        rTaskSpace = Jfull'*C*( distanceEF + Jfull*q );
    end
    
    %Regularization in State Space
    wq = 10.^[1*ones(szQ/2,1); 0.1*ones(szQ/2,1)];
    RStateSpace = alpha*diag(wq);
    %keyboard
    tmpM = Minus_pi_to_pi_States( qOld - q );
    tmp =  qOld - q ;
    diffQ = tmpM - tmp;
    qTarget = qOld + diffQ;
    rStateSpace = RStateSpace*qTarget;

    if flgActivateTSCosts
        r = rTaskSpace + rStateSpace;
        R = RTaskSpace + RStateSpace;
        %costs
        ct = dt * ( (tmpM)'*RStateSpace*(tmpM) + distanceEF'*C*distanceEF + u'*H*u ); 
    else
        r = rStateSpace;
        R = RStateSpace;
        %costs
        ct = dt * ( (tmpM)'*RStateSpace*(tmpM) + u'*H*u ); 
    end
     
end
 
function [ numEF, EF, J ] = cFnJacobianPendulum( q, l )

    numEF   = 1;
    EF      = [l*sin(q(1,:)) ; -l*cos(q(1,:))];
    if nargout > 2
    	J       = [l*cos(q(1,:)) ;  l*sin(q(1,:))];
    end
end

