% **********************************************************
% *
% * Robust Policy Updates for Stochastic Optimal Control
% *   E. Rueckert, M. Mindt, J. Peters and G. Neumann (HUMANOIDS 2014)
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert@ias.tu-darmstadt.de
% *
% **********************************************************
% * Demo file
% * running three experiments of planning swing-up movements in a simple pendulum, 
%   where for illustrative purposes the time step is set to 100ms. 
%   Without regularization (alpha=1e-12) the policy updates in the planner oscillate.
%   Note that in more complex robot tasks the effects of divergences or oscillations 
%   occur by simply increasing the number of objectives (see the paper).
%
%        The script generates a number of plots illustrating 
%            a) The internal optimization process of the planning method (the iteratively improved joint angle traj.)
%            b) A plot illustrating the convergence behavior of the planner for different alpha (regularization) parameters
%            c) The resulting optimal planned joint angles and torques
%        Note that the planner also returns a time varying feedback controller that can be used to control 
%        a real or a simulated robot.

clear all;
close all;

% PARAMETERS
alpha = [1e-12, 1, 10];         %regularization parameter
flgUseAnalyticDynamics = 1;     %0...pseudo-dyn, 1...analytic dyn
flgGeneratePlotsInRSOC = 1;     %1... plot illustrating a) The internal optimization process of the planning method 

% PATHS
addpath([pwd filesep 'RobotModel']);
addpath([pwd filesep 'RSOC']);

% ROBOT MODEL
model = Pendulum(flgUseAnalyticDynamics);

legendTxt = {};
for expId = 1:length(alpha)
    
    % PLANNING TASK
    task = PendulumSwingUpTask(alpha(expId), model.dt, model.l, model.n);    
    
    
    
    % PLOT STUFF
    if flgGeneratePlotsInRSOC > 0 %plot RSOC convergence details
        % Generate Movement Plan
        [l, L, q, u, costs, data] = RSOC(task, model);
        
        %Plot the internal generated plans
        internalPhis = zeros(model.n, size(data,2));
        for itId=1:size(data,2)
            internalPhis(:,itId) = data{itId}.q(1,:)';
        end
        cMap = copper(size(data,2));
        cMap = cMap(end:-1:1,:);
        hFig=figure;
        hold all;
        set(hFig,'Color','white');
        lhnd = zeros(size(data));
        for itId=1:size(data,2)
            lhnd(1,itId) = plot(internalPhis(:,itId), 'color', cMap(itId,:),'linewidth',1);
        end
        xlabel('time step');
        ylabel('phi');
        legend([lhnd(1,1), lhnd(1,end)], 'phi^{[0]}',['phi^{[' num2str(size(data,2)) ']}'], ...
            'location', 'northwest');
        title(['alpha=' num2str(alpha(expId))]);
        
    else
        % Generate Movement Plan
        [l, L, q, u, costs] = RSOC(task, model); 
    end
    
    if expId == 2 %illustrations for alpha=1 only
        hFig=figure;
        set(hFig,'Color','white');
        plot(q');
        legend('phi','velocity');
        title(['alpha=' num2str(alpha(expId))]);

        hFig=figure;
        set(hFig,'Color','white');
        plot(u');
        legend('control');
        title(['alpha=' num2str(alpha(expId))]);
    end
    
    costsPerAlpha(expId, :) = costs';
    legendTxt = [legendTxt {num2str(alpha(expId))}];
end
 

hFig=figure;
set(hFig,'Color','white');
semilogy(costsPerAlpha');
legend(legendTxt{:});
xlabel('RSOC iterations');
ylabel('costs in log space');

